==> bigloo2.0a.tar.gz: bigloo2.0a/cigloo/Translate/tspec.scm <==
;*---------------------------------------------------------------------*/
;*   A pratical implementation for the Scheme programming language     */
;*                                                                     */
;*                                    ,--^,                            */
;*                              _ ___/ /|/                             */
;*                          ,;'( )__, ) '                              */
;*                         ;;  //   L__.                               */
;*                         '   \\   /  '                               */
;*                              ^   ^                                  */
;*                                                                     */
;*   Copyright (c) 1992-1999 Manuel Serrano                            */
;*                                                                     */
;*     Bug descriptions, use reports, comments or suggestions are      */
;*     welcome. Send them to                                           */
;*       bigloo-request@kaolin.unice.fr                                */
;*       http://kaolin.unice.fr/bigloo                                 */
;*                                                                     */
;*   This program is free software; you can redistribute it            */
;*   and/or modify it under the terms of the GNU General Public        */
;*   License as published by the Free Software Foundation; either      */
;*   version 2 of the License, or (at your option) any later version.  */
;*                                                                     */
;*   This program is distributed in the hope that it will be useful,   */
;*   but WITHOUT ANY WARRANTY; without even the implied warranty of    */
;*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     */
;*   GNU General Public License for more details.                      */
;*                                                                     */
;*   You should have received a copy of the GNU General Public         */
;*   License along with this program; if not, write to the Free        */
;*   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,   */
;*   MA 02111-1307, USA.                                               */
;*---------------------------------------------------------------------*/
;*=====================================================================*/
;*    serrano/prgm/project/bigloo/cigloo/Translate/tspec.scm           */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Nov 30 17:24:36 1995                          */
;*    Last change :  Thu Feb 26 13:58:20 1998 (serrano)                */
;*    -------------------------------------------------------------    */
;*    The type specification handling                                  */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module translate_tspec
   (include "Translate/ast.sch"
	    "Translate/type.sch")
   (import  engine_param
	    translate_type
	    translate_decl
	    tools_error
	    tools_speek)
   (export  (tspec->type                <tspec>)
	    (simple->type               <type> <string>)
	    (type+tspec-coumpound->type <type> <tspec>)))
   
;*---------------------------------------------------------------------*/
;*    tspec->type ...                                                  */
;*---------------------------------------------------------------------*/
(define (tspec->type tspec)
   (cond
      ((default-tspec? tspec)
       (tspec->default-type))
      ((simple-tspec? tspec)
       (tspec-simple->type (car tspec)))
      (else
       (tspec-coumpound->type tspec))))

;*---------------------------------------------------------------------*/
;*    default-tspec? ...                                               */
;*---------------------------------------------------------------------*/
(define (default-tspec? tspec)
   (null? tspec))

;*---------------------------------------------------------------------*/
;*    simple-tspec? ...                                                */
;*---------------------------------------------------------------------*/
(define (simple-tspec? tspec)
   (and (pair? tspec) (null? (cdr tspec))))

;*---------------------------------------------------------------------*/
;*    tspec->default-type ...                                          */
;*---------------------------------------------------------------------*/
(define (tspec->default-type)
   (simple->type 'int "int"))
   
;*---------------------------------------------------------------------*/
;*    tspec-simple->type ...                                           */
;*---------------------------------------------------------------------*/
(define (tspec-simple->type tspec)
   (case (type-spec-class tspec)
      ((struct)
       (struct->type tspec))
      ((enum)
       (enum->type tspec))
      (else
       (simple->type (type-spec-class tspec) (type-spec-c-name tspec)))))
 
;*---------------------------------------------------------------------*/
;*    simple->type ...                                                 */
;*---------------------------------------------------------------------*/
(define (simple->type class c-name)
   (let ((cell (assq class *c-type-alist*)))
      (if (pair? cell)
	  (make-simple (cdr cell) c-name (list class))
	  (make-simple c-name c-name (list c-name)))))

;*---------------------------------------------------------------------*/
;*    struct->type ...                                                 */
;*---------------------------------------------------------------------*/
(define (struct->type tspec)
   (verbose 2 "struct->type: " tspec #\Newline)
   (let* ((sspec   (type-spec-value tspec))
	  (s-ident (struct-spec-id sspec))
	  (s-id    (ident-id s-ident))
	  (b-id    (string-append "s-" s-id))
	  (c-name  (string-append (if (eq? (struct-spec-class sspec) 'struct)
				      "struct "
				      "union ")
				  s-id))
	  (struct  (make-cstruct b-id
				 c-name
				 s-id
				 (struct-spec-class sspec)
				 '()))
	  (fields  (let loop ((fields (struct-spec-fields sspec))
			      (res    '()))
		      (if (null? fields)
			  (reverse! res)
			  (let* ((field (car fields))
				 (tspec (car field))
				 (decls (cadr field)))
			     (let liip ((decls decls)
					(res   res))
				(cond
				   ((null? decls)
				    (loop (cdr fields) res))
				   ((and (not (decl-a-ptr (car decls)))
					 (not (decl-a-decl2 (car decls))))
				    ;; we skip bitfield with no declarator.
				    (loop (cdr fields) res))
				   (else
				    (let* ((decl    (car decls))
					   (type    (tspec->type tspec))
					   (type    (type+decl->type type
								     decl))
					   (f-ident (get-decl-ident decl))
					   (f-id    (ident-id f-ident)))
				       (liip (cdr decls)
					     (cons (cons type f-id)
						   res)))))))))))
      (if (pair? fields)
	  (struct-t-fields-set! struct fields))
      struct))

;*---------------------------------------------------------------------*/
;*    enum->type ...                                                   */
;*---------------------------------------------------------------------*/
(define (enum->type tspec)
   (let* ((espec   (type-spec-value tspec))
	  (e-ident (enum-spec-id espec))
	  (e-id    (if (ident? e-ident)
		       (ident-id e-ident)
		       (symbol->string
			(gensym (string-append *iname* "__e")))))
	  (b-id    (string-append "e-" e-id))
	  (c-name  (if (ident? e-ident)
		       (string-append "enum " e-id)
		       "long"))
	  (enum    (make-enum e-id c-name b-id '())))
      (if (pair? (enum-spec-enumerator-list espec))
	  (enum-t-fields-set! enum (enum-spec-enumerator-list espec)))
      enum))

;*---------------------------------------------------------------------*/
;*    tspec-coumpound->type ...                                        */
;*---------------------------------------------------------------------*/
(define (tspec-coumpound->type tspecs)
   (tspec-coumpound
    tspecs
    (lambda (unsigned signed long longlong short tspec)
       (let ((tspec (cond
		       (longlong (begin
				    (type-spec-class-set!  longlong 'longlong)
				    (type-spec-c-name-set! longlong "long long")
				    longlong))
		       (long     long)
		       (short    short)
		       (else     tspec))))
	  (if (not tspec)
	      (error/ast 'type "Illegal type" (car tspecs))
	      (let ((type (type-spec-class tspec)))
		 (cond
		    (unsigned
		     (let ((cell (assq type *c-unsigned-type-alist*)))
			(if (not (pair? cell))
			    (error/ast 'unsigned "Illegal type" unsigned)
			    (let ((name (string-append
					 "unsigned "
					 (type-spec-c-name tspec))))
			       (make-simple (cdr cell)
					    name
					    (list 'unsigned type))))))
		    (signed
		     (let ((cell (assq type *c-signed-type-alist*)))
			(if (not (pair? cell))
			    (error/ast 'signed "Illegal type" signed)
			    (let ((name (string-append
					 "signed "
					 (type-spec-c-name tspec))))
			       (make-simple (cdr cell)
					    name
					    (list 'signed type))))))
		    (else
		     (tspec-simple->type tspec)))))))))

;*---------------------------------------------------------------------*/
;*    type+tspec-coumpound->type ...                                   */
;*---------------------------------------------------------------------*/
(define (type+tspec-coumpound->type type tspecs)
   (tspec-coumpound
    tspecs
    (lambda (unsigned signed long longlong short tspec)
       (let ((tspec (cond
		       (longlong (simple->type 'longlong "longlong"))
		       (long     long)
		       (short    short)
		       (else     tspec))))
	  (if (not tspec)
	      type
	      (error/ast 'type "Illegal type" (car tspecs)))))))

;*---------------------------------------------------------------------*/
;*    tspec-coumpound ...                                              */
;*---------------------------------------------------------------------*/
(define (tspec-coumpound tspecs fun)
   (let loop ((unsigned #f)
	      (signed   #f)
	      (volatile #f)
	      (const    #f)
	      (long     #f)
	      (longlong #f)
	      (short    #f)
	      (tspec    #f)
	      (tspcs   tspecs))
      (cond
	 ((null? tspcs)
	  (fun unsigned signed long longlong short tspec))
	 (else
	  (let ((ts (car tspcs)))
	     (case (type-spec-class ts)
		((unsigned)
		 (if unsigned
		     (error/ast 'unsigned "Duplicate type specification" ts)
		     (loop ts
			   signed
			   volatile
			   const
			   long
			   longlong
			   short
			   tspec
			   (cdr tspcs))))
		((signed)
		 (if signed
		     (error/ast 'signed "Duplicate type specification" ts)
		     (loop unsigned
			   ts
			   volatile
			   const
			   long
			   longlong
			   short
			   tspec
			   (cdr tspcs))))
		((volatile)
		 (if volatile
		     (error/ast 'volatile "Duplicate type specification" ts)
		     (loop unsigned
			   signed
			   ts
			   const
			   long
			   longlong
			   short
			   tspec
			   (cdr tspcs))))
		((const)
		 (if const
		     (error/ast 'const "Duplicate type specification" ts)
		     (loop unsigned
			   signed
			   volatile
			   ts
			   long
			   longlong
			   short
			   tspec
			   (cdr tspcs))))
		(else
		 (cond
		    (tspec
		     (error/ast 'type "Illegal type" ts))
		    ((eq? (type-spec-class ts) 'short)
		     (if (and (not long)
			      (not longlong)
			      (not short)
			      (not tspec))
			 (loop unsigned
			       signed
			       volatile
			       const
			       #f
			       #f
			       ts
			       #f
			       (cdr tspcs))
			 (error/ast 'type "Illegal type" ts)))
		    ((eq? (type-spec-class ts) 'long)
		     (cond
			((and (not longlong)
			      (not short)
			      (not tspec))
			 (if long
			     (loop unsigned
				   signed
				   volatile
				   const
				   #f
				   ts
				   #f
				   #f
				   (cdr tspcs))
			     (loop unsigned
				   signed
				   volatile
				   const
				   ts
				   #f
				   #f
				   #f
				   (cdr tspcs))))
			(else
			 (error/ast 'type "Illegal type" ts))))
		    ((eq? (type-spec-class ts) 'int)
		     (cond
			(short
			 (loop unsigned
			       signed
			       volatile
			       const
			       short
			       #f
			       #f
			       short
			       (cdr tspcs)))
			(long
			 (loop unsigned
			       signed
			       volatile
			       const
			       long
			       longlong
			       #f
			       long
			       (cdr tspcs)))
			(else
			 (loop unsigned
			       signed
			       volatile
			       const
			       #f
			       #f
			       #f
			       ts
			       (cdr tspcs)))))
		    (else
		     (loop unsigned
			   signed
			   volatile
			   const
			   long
			   longlong
			   short
			   ts
			   (cdr tspcs)))))))))))

==> bigloo_2.4b.orig.tar.gz: bigloo2.4b/cigloo/Translate/tspec.scm <==
;*=====================================================================*/
;*    serrano/prgm/project/bigloo/cigloo/Translate/tspec.scm           */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Nov 30 17:24:36 1995                          */
;*    Last change :  Mon Dec  3 18:19:40 2001 (serrano)                */
;*    -------------------------------------------------------------    */
;*    The type specification handling                                  */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module translate_tspec
   (include "Translate/ast.sch"
	    "Translate/type.sch")
   (import  engine_param
	    translate_type
	    translate_decl
	    parser_parser
	    tools_error
	    tools_speek)
   (export  (tspec->type                <tspec>)
	    (simple->type               <type> <string>)
	    (type+tspec-coumpound->type <type> <tspec>)))
   
;*---------------------------------------------------------------------*/
;*    tspec->type ...                                                  */
;*---------------------------------------------------------------------*/
(define (tspec->type tspec)
   (cond
      ((default-tspec? tspec)
       (tspec->default-type))
      ((simple-tspec? tspec)
       (tspec-simple->type (car tspec)))
      (else
       (tspec-coumpound->type tspec))))

;*---------------------------------------------------------------------*/
;*    default-tspec? ...                                               */
;*---------------------------------------------------------------------*/
(define (default-tspec? tspec)
   (null? tspec))

;*---------------------------------------------------------------------*/
;*    simple-tspec? ...                                                */
;*---------------------------------------------------------------------*/
(define (simple-tspec? tspec)
   (and (pair? tspec) (null? (cdr tspec))))

;*---------------------------------------------------------------------*/
;*    tspec->default-type ...                                          */
;*---------------------------------------------------------------------*/
(define (tspec->default-type)
   (simple->type 'int "int"))
   
;*---------------------------------------------------------------------*/
;*    tspec-simple->type ...                                           */
;*---------------------------------------------------------------------*/
(define (tspec-simple->type tspec)
   (case (type-spec-class tspec)
      ((struct)
       (struct->type tspec))
      ((enum)
       (enum->type tspec))
      (else
       (simple->type (type-spec-class tspec) (type-spec-c-name tspec)))))
 
;*---------------------------------------------------------------------*/
;*    simple->type ...                                                 */
;*---------------------------------------------------------------------*/
(define (simple->type class c-name)
   (let ((cell (assq class *c-type-alist*)))
      (if (pair? cell)
	  (make-simple (cdr cell) c-name (list class))
	  (make-simple c-name c-name (list c-name)))))

;*---------------------------------------------------------------------*/
;*    struct->type ...                                                 */
;*---------------------------------------------------------------------*/
(define (struct->type tspec)
   (verbose 2 "struct->type: " tspec #\Newline)
   (let* ((sspec   (type-spec-value tspec))
	  ;; Look for an alias for this type.  It will only be present if
	  ;; the struct was anonymous and inside a typedef.
	  (alias-pair   (assq sspec *anonymous-struct-alist*))
	  (alias   (and alias-pair 
			(not (null? (cdr alias-pair)))
			(cdr alias-pair)))
	  (s-ident (struct-spec-id sspec))
	  (s-id    (ident-id s-ident))
	  (b-id    (string-append "s-" s-id))
	  (c-name  (cond 
		    ;; Use the typedef as the c-name if available.
		    (alias
		     (ident-id alias))
		    ;; If this is an anonymous struct (or union)
		    ;; without an alias, it has no useful c-name.  It
		    ;; was probably embedded in another struct.
		    (alias-pair
		     "void")
		    (else
		       (string-append
			(if (eq? (struct-spec-class sspec) 'struct)
			    "struct "
			    "union ")
			s-id))))
	  (struct  (make-cstruct b-id
				 c-name
				 s-id
				 (struct-spec-class sspec)
				 '()))
	  (fields  (let loop ((fields (struct-spec-fields sspec))
			      (res    '()))
		      (if (null? fields)
			  (reverse! res)
			  (let* ((field (car fields))
				 (tspec (car field))
				 (decls (cadr field)))
			     (let liip ((decls decls)
					(res   res))
				(cond
				   ((null? decls)
				    (loop (cdr fields) res))
				   ((and (not (decl-a-ptr (car decls)))
					 (not (decl-a-decl2 (car decls))))
				    ;; we skip bitfield with no declarator.
				    (loop (cdr fields) res))
				   (else
				    (let* ((decl    (car decls))
					   (type    (tspec->type tspec))
					   (type    (type+decl->type type
								     decl))
					   (f-ident (get-decl-ident decl))
					   (f-id    (ident-id f-ident)))
				       (liip (cdr decls)
					     (cons (cons type f-id)
						   res)))))))))))
      (if (pair? fields)
	  (struct-t-fields-set! struct fields))
      struct))

;*---------------------------------------------------------------------*/
;*    enum->type ...                                                   */
;*---------------------------------------------------------------------*/
(define (enum->type tspec)
   (let* ((espec   (type-spec-value tspec))
	  (e-ident (enum-spec-id espec))
	  (e-id    (if (ident? e-ident)
		       (ident-id e-ident)
		       (symbol->string
			(gensym (string-append *iname* "__e")))))
	  (b-id    (string-append "e-" e-id))
	  (c-name  (if (ident? e-ident)
		       (string-append "enum " e-id)
		       "long"))
	  (enum    (make-enum e-id c-name b-id '())))
      (if (pair? (enum-spec-enumerator-list espec))
	  (enum-t-fields-set! enum (enum-spec-enumerator-list espec)))
      enum))

;*---------------------------------------------------------------------*/
;*    tspec-coumpound->type ...                                        */
;*---------------------------------------------------------------------*/
(define (tspec-coumpound->type tspecs)
   (tspec-coumpound
    tspecs
    (lambda (unsigned signed long longlong short tspec)
       (let ((tspec (cond
		       (longlong (begin
				    (type-spec-class-set!  longlong 'longlong)
				    (type-spec-c-name-set! longlong "long long")
				    longlong))
		       (long     long)
		       (short    short)
		       (else     tspec))))
	  (if (not tspec)
	      (error/ast 'type1 "Illegal type" (car tspecs))
	      (let ((type (type-spec-class tspec)))
		 (cond
		    (unsigned
		     (let ((cell (assq type *c-unsigned-type-alist*)))
			(if (not (pair? cell))
			    (error/ast 'unsigned "Illegal type" unsigned)
			    (let ((name (string-append
					 "unsigned "
					 (type-spec-c-name tspec))))
			       (make-simple (cdr cell)
					    name
					    (list 'unsigned type))))))
		    (signed
		     (let ((cell (assq type *c-signed-type-alist*)))
			(if (not (pair? cell))
			    (error/ast 'signed "Illegal type" signed)
			    (let ((name (string-append
					 "signed "
					 (type-spec-c-name tspec))))
			       (make-simple (cdr cell)
					    name
					    (list 'signed type))))))
		    (else
		     (tspec-simple->type tspec)))))))))

;*---------------------------------------------------------------------*/
;*    type+tspec-coumpound->type ...                                   */
;*---------------------------------------------------------------------*/
(define (type+tspec-coumpound->type type tspecs)
   (tspec-coumpound
    tspecs
    (lambda (unsigned signed long longlong short tspec)
       (let ((tspec (cond
		       (longlong (simple->type 'longlong "longlong"))
		       (long     long)
		       (short    short)
		       (else     tspec))))
	  (if (not tspec)
	      type
	      (error/ast 'type2 "Illegal type" (car tspecs)))))))

;*---------------------------------------------------------------------*/
;*    tspec-coumpound ...                                              */
;*---------------------------------------------------------------------*/
(define (tspec-coumpound tspecs fun)
   (let loop ((unsigned #f)
	      (signed   #f)
	      (volatile #f)
	      (const    #f)
	      (long     #f)
	      (longlong #f)
	      (short    #f)
	      (tspec    #f)
	      (tspcs   tspecs))
      (cond
	 ((null? tspcs)
	  (fun unsigned signed long longlong short tspec))
	 (else
	  (let ((ts (car tspcs)))
	     (case (type-spec-class ts)
		((unsigned)
		 (if unsigned
		     (error/ast 'unsigned "Duplicate type specification" ts)
		     (loop ts
			   signed
			   volatile
			   const
			   long
			   longlong
			   short
			   tspec
			   (cdr tspcs))))
		((signed)
		 (if signed
		     (error/ast 'signed "Duplicate type specification" ts)
		     (loop unsigned
			   ts
			   volatile
			   const
			   long
			   longlong
			   short
			   tspec
			   (cdr tspcs))))
		((volatile)
		 (if volatile
		     (error/ast 'volatile "Duplicate type specification" ts)
		     (loop unsigned
			   signed
			   ts
			   const
			   long
			   longlong
			   short
			   tspec
			   (cdr tspcs))))
		((const)
		 (if const
		     (error/ast 'const "Duplicate type specification" ts)
		     (loop unsigned
			   signed
			   volatile
			   ts
			   long
			   longlong
			   short
			   tspec
			   (cdr tspcs))))
		(else
		 (cond
		    (tspec
		     (error/ast (type-spec-class ts) "Illegal type" ts))
		    ((eq? (type-spec-class ts) 'short)
		     (if (and (not long)
			      (not longlong)
			      (not short)
			      (not tspec))
			 (loop unsigned
			       signed
			       volatile
			       const
			       #f
			       #f
			       ts
			       #f
			       (cdr tspcs))
			 (error/ast 'type4 "Illegal type" ts)))
		    ((eq? (type-spec-class ts) 'long)
		     (cond
			((and (not longlong)
			      (not short)
			      (not tspec))
			 (if long
			     (loop unsigned
				   signed
				   volatile
				   const
				   #f
				   ts
				   #f
				   #f
				   (cdr tspcs))
			     (loop unsigned
				   signed
				   volatile
				   const
				   ts
				   #f
				   #f
				   #f
				   (cdr tspcs))))
			(else
			 (error/ast 'type5 "Illegal type" ts))))
		    ((eq? (type-spec-class ts) 'int)
		     (cond
			(short
			 (loop unsigned
			       signed
			       volatile
			       const
			       short
			       #f
			       #f
			       short
			       (cdr tspcs)))
			(long
			 (loop unsigned
			       signed
			       volatile
			       const
			       long
			       longlong
			       #f
			       long
			       (cdr tspcs)))
			(else
			 (loop unsigned
			       signed
			       volatile
			       const
			       #f
			       #f
			       #f
			       ts
			       (cdr tspcs)))))
		    (else
		     (loop unsigned
			   signed
			   volatile
			   const
			   long
			   longlong
			   short
			   ts
			   (cdr tspcs)))))))))))
